@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* input.h: functions for input of command line */
-/* (C) 2007-2011 magicant */
+/* (C) 2007-2018 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -60,15 +60,20 @@ extern inputresult_T read_input(
     __attribute__((nonnull));
 
 /* The type of input functions.
- * Input is done line-wise: one line is read at a time.
- * The input string is terminated by newline L'\n' except when there is no
- * newline at the end of input.
+ * An input function reads input and appends it to buffer `buf'.
+ * Input is done line-wise: the buffer contents are always terminated by a
+ * newline character (L'\n') except when the end of file is reached and the last
+ * line does not have a newline.
+ * An input function should not read more than one line at a time, as commands
+ * (which may contain alias definitions) should be executed as soon as possible,
+ * before the next line is parsed.
  * The result is indicated by a value of the `inputresult_T' type. If the return
  * value is other than INPUT_OK, the buffer is unchanged.
- * The input function may be called even after it returned a non-zero. */
+ * The input function may be called even after it returned a value other than
+ * INPUT_OK. */
 typedef inputresult_T inputfunc_T(struct xwcsbuf_T *buf, void *inputinfo);
 
-/* input functions to be used as `inputfunc_T' */
+/* input functions */
 extern inputresult_T input_wcs(struct xwcsbuf_T *buf, void *inputinfo)
     __attribute__((nonnull));
 extern inputresult_T input_file(struct xwcsbuf_T *buf, void *inputinfo)
@@ -94,9 +99,6 @@ struct input_file_info_T {
 struct input_interactive_info_T {
     struct input_file_info_T *fileinfo;
     int prompttype;
-#if YASH_ENABLE_LINEEDIT
-    wchar_t *linebuffer;
-#endif
 };
 
 
