@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* redir.c: manages file descriptors and provides functions for redirections */
-/* (C) 2007-2018 magicant */
+/* (C) 2007-2020 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -117,9 +117,6 @@ bool write_all(int fd, const void *data, size_t size)
 static void reset_shellfdmin(void);
 
 
-/* True iff the standard input is redirected */
-static bool is_stdin_redirected = false;
-
 /* Set of file descriptors used by the shell.
  * These file descriptors cannot be used by the user. */
 static fd_set shellfds;
@@ -291,7 +288,6 @@ struct savefd_T {
     struct savefd_T *next;
     int  sf_origfd;            /* original file descriptor */
     int  sf_copyfd;            /* copied file descriptor */
-    bool sf_stdin_redirected;  /* original `is_stdin_redirected' */
 };
 
 static char *expand_redir_filename(const struct wordunit_T *filename)
@@ -315,9 +311,9 @@ static int open_process_redirection(const embedcmd_T *command, redirtype_T type)
     __attribute__((nonnull));
 
 
-/* Opens redirection.
- * If `save' is non-NULL, the original FD is saved and a pointer to the info is
- * assigned to `*save' (whether successful or not).
+/* Opens redirections.
+ * The original FDs are saved and a pointer to the restoration info is assigned
+ * to `*save' (whether successful or not).
  * Returns true iff successful. */
 bool open_redirections(const redir_T *r, savefd_T **save)
 {
@@ -433,9 +429,6 @@ openwithflags:
 	    }
 	}
 
-	if (r->rd_fd == STDIN_FILENO)
-	    is_stdin_redirected = true;
-
 	r = r->next;
     }
     return true;
@@ -446,10 +439,9 @@ openwithflags:
 char *expand_redir_filename(const struct wordunit_T *filename)
 {
     if (is_interactive) {
-	return expand_single_with_glob(filename, TT_SINGLE);
+	return expand_single_with_glob(filename);
     } else {
-	wchar_t *result = expand_single_and_unescape(
-		filename, TT_SINGLE, true, false);
+	wchar_t *result = expand_single(filename, TT_SINGLE, Q_WORD, ES_NONE);
 	if (result == NULL)
 	    return NULL;
 	char *mbsresult = realloc_wcstombs(result);
@@ -476,7 +468,6 @@ void save_fd(int fd, savefd_T **save)
     s->next = *save;
     s->sf_origfd = fd;
     s->sf_copyfd = copyfd;
-    s->sf_stdin_redirected = is_stdin_redirected;
     *save = s;
 }
 
@@ -758,8 +749,7 @@ error:
  * temporary file. */
 int open_heredocument(const wordunit_T *contents)
 {
-    wchar_t *wcontents = expand_single_and_unescape(
-	    contents, TT_NONE, false, false);
+    wchar_t *wcontents = expand_single(contents, TT_NONE, Q_INDQ, ES_NONE);
     if (wcontents == NULL)
 	return -1;
 
@@ -900,7 +890,6 @@ void undo_redirections(savefd_T *save)
 	} else {
 	    xclose(save->sf_origfd);
 	}
-	is_stdin_redirected = save->sf_stdin_redirected;
 
 	savefd_T *next = save->next;
 	free(save);
@@ -928,21 +917,18 @@ void clear_savefd(savefd_T *save)
  * standard input is not yet redirected. */
 void maybe_redirect_stdin_to_devnull(void)
 {
-    int fd;
-
-    if (do_job_control || is_stdin_redirected)
+    if (do_job_control)
 	return;
 
     if (xclose(STDIN_FILENO) < 0)
 	return;
 
-    fd = open("/dev/null", O_RDONLY);
+    int fd = open("/dev/null", O_RDONLY);
     if (fd < 0) {
 	//xerror(errno, Ngt("cannot redirect the standard input to /dev/null"));
     } else {
 	assert(fd == STDIN_FILENO);
     }
-    is_stdin_redirected = true;
 }
 
 
