@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* util.h: miscellaneous utility functions */
-/* (C) 2007-2012 magicant */
+/* (C) 2007-2021 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -135,8 +135,20 @@ void *xmallocs(size_t mainsize, size_t count, size_t elemsize)
 /* Attempts `realloc' and aborts the program on failure. */
 void *xrealloc(void *ptr, size_t size)
 {
+    /* The behavior of `realloc(non_null_pointer, 0)' is unreliable. Some
+     * implementations free the previously allocated region and return NULL.
+     * Some reallocate an empty region and return a pointer to it. The latter
+     * may fail to allocate the new region, leaving the previous region intact
+     * and returning NULL. That means, when `realloc(non_null_pointer, 0)'
+     * returned NULL, we cannot tell if the previous region has been freed or
+     * not. */
+    if (size == 0) {
+	free(ptr);
+	return NULL;
+    }
+
     void *result = realloc(ptr, size);
-    if (result == NULL && size > 0)
+    if (result == NULL)
 	alloc_failed();
     return result;
 }
@@ -165,18 +177,24 @@ void *xreallocs(void *ptr, size_t mainsize, size_t count, size_t elemsize)
 
 /********** String Utilities **********/
 
+#if !HAVE_STRNLEN
 extern size_t xstrnlen(const char *s, size_t maxlen)
     __attribute__((pure,nonnull));
-extern char *xstrndup(const char *s, size_t maxlen)
-    __attribute__((malloc,warn_unused_result,nonnull));
-static inline char *xstrdup(const char *s)
+#endif
+#if !HAVE_STRDUP
+extern char *xstrdup(const char *s)
     __attribute__((malloc,warn_unused_result,nonnull));
+#endif
+#if !HAVE_WCSNLEN
 extern size_t xwcsnlen(const wchar_t *s, size_t maxlen)
     __attribute__((pure,nonnull));
+#endif
 extern wchar_t *xwcsndup(const wchar_t *s, size_t maxlen)
     __attribute__((malloc,warn_unused_result,nonnull));
+#if !HAVE_WCSDUP
 static inline wchar_t *xwcsdup(const wchar_t *s)
     __attribute__((malloc,warn_unused_result,nonnull));
+#endif
 extern _Bool xstrtoi(const char *s, int base, int *resultp)
     __attribute__((warn_unused_result,nonnull));
 extern _Bool xwcstoi(const wchar_t *s, int base, int *resultp)
@@ -196,28 +214,36 @@ extern void *copyaswcs(const void *p)
 # ifndef strnlen
 extern size_t strnlen(const char *s, size_t maxlen);
 # endif
-# define xstrnlen(s,maxlen) strnlen(s,maxlen)
+# define xstrnlen strnlen
 #endif
+
+#if HAVE_STRDUP
+# ifndef strdup
+extern char *strdup(const char *s);
+# endif
+# define xstrdup strdup
+#endif
+
 #if HAVE_WCSNLEN
 # ifndef wcsnlen
 extern size_t wcsnlen(const wchar_t *s, size_t maxlen);
 # endif
-# define xwcsnlen(s,maxlen) wcsnlen(s,maxlen)
+# define xwcsnlen wcsnlen
 #endif
 
-/* Returns a newly malloced copy of the specified string.
- * Aborts the program if failed to allocate memory. */
-char *xstrdup(const char *s)
-{
-    return xstrndup(s, Size_max);
-}
-
+#if HAVE_WCSDUP
+# ifndef wcsdup
+extern wchar_t *wcsdup(const wchar_t *s);
+# endif
+# define xwcsdup wcsdup
+#else
 /* Returns a newly malloced copy of the specified string.
  * Aborts the program if failed to allocate memory. */
 wchar_t *xwcsdup(const wchar_t *s)
 {
     return xwcsndup(s, Size_max);
 }
+#endif
 
 
 /* These macros are used to cast the argument properly.
